package com.zehfernando.ld27.display.scenes.game.world {
	import nape.callbacks.CbType;
	import nape.geom.Vec2;
	import nape.phys.Body;
	import nape.phys.BodyType;
	import nape.shape.Polygon;
	import nape.space.Space;

	import com.zehfernando.input.binding.KeyActionBinder;
	import com.zehfernando.ld27.display.scenes.game.actors.Actor;
	import com.zehfernando.ld27.display.scenes.game.actors.Background;
	import com.zehfernando.ld27.display.scenes.game.actors.Building;
	import com.zehfernando.ld27.display.scenes.game.actors.Cloud;
	import com.zehfernando.ld27.display.scenes.game.actors.Ground;
	import com.zehfernando.ld27.display.scenes.game.actors.Human;
	import com.zehfernando.ld27.display.scenes.game.actors.Player;
	import com.zehfernando.utils.AppUtils;
	import com.zehfernando.utils.MathUtils;
	import com.zehfernando.utils.RandomGenerator;
	import com.zehfernando.utils.console.log;
	/**
	 * @author zeh fernando
	 */
	public class WorldModel {

		// Constants
		public static const EDGE_THICKNESS:Number = 2;

		public static const COLLISION_GROUP_PLAYER:int = 2;
		public static const COLLISION_GROUP_PARTICLES:int = 4;
		public static const COLLISION_GROUP_HUMANS:int = 8;
		public static const COLLISION_GROUP_LANDABLE:int = 16;
		public static const COLLISION_GROUP_BEAM:int = 32;

		// CBTypes
		public static var CB_TYPE_BEAM:CbType = new CbType();
		public static var CB_TYPE_BEAMABLE:CbType = new CbType();

		// Properties
		private var cameraX:Number;										// Position in the world of what's at the center of the view
		private var cameraY:Number;
		private var worldWidth:Number;
		private var worldHeight:Number;
		private var groundHeight:Number;

		private var levelData:Object;

		private var buildings:Vector.<Building>;

		// Instances
		private var binder:KeyActionBinder;
		private var space:Space;
		private var actors:Vector.<Actor>;
		private var playerActor:Player;
		private var worldView:WorldView;
		private var worldHUD:WorldHUD;
		private var targetObjective:int;

		private var _gameIsEnded:Boolean;
		private var _playerWon:Boolean;
		private var _playerLost:Boolean;

		private var timeGameEnded:Number;


		// ================================================================================================================
		// CONSTRUCTOR ----------------------------------------------------------------------------------------------------

		public function WorldModel(__levelData:Object, __worldView:WorldView, __worldHUD:WorldHUD) {
			actors = new Vector.<Actor>();

			cameraX = 0;
			cameraY = 0;
			groundHeight = 100;

			levelData = __levelData;
			worldView = __worldView;
			worldHUD = __worldHUD;

			_playerWon = false;
			_playerLost = false;
			timeGameEnded = NaN;

			// Create key binder
			binder = new KeyActionBinder(AppUtils.getStage());

			// Create world
			var i:int;
			var gravity:Vec2 = new Vec2(0, 100);
			space = new Space(gravity);

			targetObjective = __levelData["objective"];
			worldWidth = __levelData["width"];
			worldHeight = __levelData["height"];

			// Player
			playerActor = new Player(__levelData["playerStart"]["x"], __levelData["playerStart"]["y"]);
			addActor(playerActor);

			// Boundaries
			var polygonBody:Body;
			var wallWidth:Number = 10; // In pixels;

			// Left
			polygonBody = new Body(BodyType.STATIC);
			polygonBody.position.setxy(-wallWidth, 0);
			polygonBody.shapes.add(new Polygon(Polygon.rect(0, 0, wallWidth, worldHeight)));
			polygonBody.userData["name"] = "bounds-left";
			space.bodies.add(polygonBody);
			// Right
			polygonBody = new Body(BodyType.STATIC);
			polygonBody.position.setxy(worldWidth, 0);
			polygonBody.shapes.add(new Polygon(Polygon.rect(0, 0, wallWidth, worldHeight)));
			polygonBody.userData["name"] = "bounds-right";
			space.bodies.add(polygonBody);
			// Top
			polygonBody = new Body(BodyType.STATIC);
			polygonBody.position.setxy(0, -wallWidth);
			polygonBody.shapes.add(new Polygon(Polygon.rect(0, 0, worldWidth, wallWidth)));
			polygonBody.userData["name"] = "bounds-top";
			space.bodies.add(polygonBody);
			// Bottom
			polygonBody = new Body(BodyType.STATIC);
			polygonBody.position.setxy(0, worldHeight);
			polygonBody.shapes.add(new Polygon(Polygon.rect(0, 0, worldWidth, wallWidth)));
			polygonBody.userData["name"] = "bounds-bottom";
			space.bodies.add(polygonBody);

			// Add special actors

			// Background
			var background:Background = new Background(__levelData["background"], worldWidth, worldHeight);
			addActor(background);

			// Buildings
			createBuildings();

			// Ground
			var ground:Ground = new Ground(__levelData["ground"], worldWidth, groundHeight);
			addActor(ground);

			// Clouds
			var cloudCount:int = __levelData["clouds"]["count"];
			for (i = 0; i < cloudCount; i++) {
				addActor(new Cloud(RandomGenerator.getInRange(100, worldWidth - 100), MathUtils.map(i, 0, cloudCount-1, __levelData["clouds"]["minY"], __levelData["clouds"]["maxY"]) + RandomGenerator.getInRange(-20, 20), MathUtils.map(i, 0, cloudCount-1, 0.5, 1), RandomGenerator.getInRange(-20, 20)));
			}

			// Humans
			createHumans();
		}


		// ================================================================================================================
		// INTERNAL INTERFACE ---------------------------------------------------------------------------------------------

		private function createBuildings():void {
			// Distributes buildings on the level in a better way than just random distribution

			var buildingCount:int = levelData["buildings"]["count"];
			var i:int;

			// Do random distribution of gap weights
			var gaps:Vector.<Number> = new Vector.<Number>();
			var totalGapWeights:Number = 0;
			for (i = 0; i < buildingCount+1; i++) {
				gaps.push(Math.random());
				totalGapWeights += gaps[i];
			}

			// Find the actual width used by buildings
			var sideMargins:Number = Building.ASSET_WIDTH; // How much should be free on the sides

			var buildingsWidth:Number = buildingCount * Building.ASSET_WIDTH;
			var availableWidth:Number = worldWidth - sideMargins * 2 - buildingsWidth;

			var randomizedGapAmount:Number = 0.2; // 0 = none (even distribution) to 1 = totally random

			// Finally, create all buildings
			var posX:Number = sideMargins;
			var building:Building;
			buildings = new Vector.<Building>();

			for (i = 0; i < buildingCount; i++) {
				posX += ((availableWidth * (1-randomizedGapAmount)) / gaps.length) + (availableWidth * randomizedGapAmount) * (gaps[i] / totalGapWeights);
				building = new Building(posX, worldHeight - groundHeight, Math.round(RandomGenerator.getInRange(levelData["buildings"]["minFloors"], levelData["buildings"]["maxFloors"])));
				buildings.push(building);
				addActor(building);
				posX += Building.ASSET_WIDTH;
			}
		}

		private function createHumans():void {
			// Create humans, finding the correct floor height for them
			var humanCount:int = levelData["humans"];
			var human:Human;
			var humanX:Number, humanY:Number;
			var i:int, j:int;
			for (i = 0; i < humanCount; i++) {
				humanX = RandomGenerator.getInRange(20, worldWidth - 20);
				humanY = worldHeight - groundHeight;

				// Find if there's any building at that position
				for (j = 0; j < buildings.length; j++) {
					if (buildings[j].x <= humanX && buildings[j].x + Building.ASSET_WIDTH >= humanX) {
						humanY = worldHeight - groundHeight - buildings[j].height;
						break;
					}
				}

				human = new Human(humanX, humanY, Human.VELOCITY_WALK * (Math.random() > 0.5 ? -1 : 1));
				addActor(human);
				humanCount--;
			}
		}


		// ================================================================================================================
		// PUBLIC INTERFACE -----------------------------------------------------------------------------------------------

		public function addActor(__actor:Actor):void {
			actors.push(__actor);
			__actor.setWorldModel(this);
			__actor.start();
		}

		public function update(__currentTimeSeconds:Number, __deltaTimeSeconds:Number):void {
			// Ran 60 times a second
			// Go through body list and update sprite positions/rotations
			space.step(__deltaTimeSeconds, 3, 8);

			var i:int;

			// Update all actors
			for (i = 0; i < actors.length; i++) {
				actors[i].update(__currentTimeSeconds, __deltaTimeSeconds);
			}

			// Remove dead actors
			for (i = 0; i < actors.length; i++) {
				if (actors[i].isDead) {
					actors[i].stop();
					actors.splice(i, 1);
					i--;
				}
			}
		}

		public function draw(__currentTimeSeconds:Number, __deltaTimeSeconds:Number):void {
			// Ran once a frame

			// Ask all the actors to update their containers as needed
			// TODO: move this somewhere else?
			for (var i:int = 0; i < actors.length; i++) {
				actors[i].draw(__currentTimeSeconds, __deltaTimeSeconds);
			}

			// Clamps camera so it won't show outside the world
			var cx:Number = Math.min(Math.max(worldView.width / 2, cameraX), worldWidth - worldView.width/2);
			var cy:Number = Math.min(Math.max(worldView.height / 2, cameraY), worldHeight - worldView.height/2);

			// Draw actual level
			worldView.draw(cx, cy, space);

			// Update HUD
			worldHUD.setFuel(playerActor.fuel, Player.MAX_FUEL);
			worldHUD.setCapturedTargets(playerActor.numCapturedTargets, targetObjective);

			if (playerActor.numCapturedTargets > targetObjective && !_playerLost) {
				// Waits a few seconds
				_playerWon = true;
				if (isNaN(timeGameEnded)) {
					timeGameEnded = __currentTimeSeconds;
				} else {
					if (__currentTimeSeconds > timeGameEnded + 3 && !_gameIsEnded) {
						_gameIsEnded = true;
						log ('Player WON');
					}
				}
			} else {
				if (playerActor.isOutOfFuel && !_playerWon) {
					_playerLost = true;
					// Waits a few seconds
					if (isNaN(timeGameEnded)) {
						timeGameEnded = __currentTimeSeconds;
					} else {
						if (__currentTimeSeconds > timeGameEnded + 3 && !_gameIsEnded) {
							_gameIsEnded = true;
							log ("Player LOST");
						}
					}
				}
			}
		}

		public function dispose():void {
			// TODO?
		}

		// ================================================================================================================
		// ACCESSOR INTERFACE ---------------------------------------------------------------------------------------------

		public function getKeyBinder():KeyActionBinder {
			return binder;
		}

		public function setCameraPosition(__x:Number, __y:Number):void {
			cameraX = __x;
			cameraY = __y;
		}

		public function getSpace():Space {
			return space;
		}

		public function getWorldView():WorldView {
			return worldView;
		}

		public function getWorldWidth():Number {
			return worldWidth;
		}

		public function getWorldHeight():Number {
			return worldHeight;
		}

		public function getPlayerActor():Player {
			return playerActor;
		}

		public function get playerWon():Boolean {
			return _playerWon && _gameIsEnded;
		}

		public function get playerLost():Boolean {
			return _playerLost && _gameIsEnded;
		}
	}
}
